/********************************************************************************
 * Copyright (c) 2021-2023 Bayerische Motoren Werke Aktiengesellschaft (BMW AG)
 * Copyright (c) 2022-2023 Ansys, Inc.
 *
 * This program and the accompanying materials are made available under the
 * terms of the Eclipse Public License 2.0 which is available at
 * http://www.eclipse.org/legal/epl-2.0.
 *
 * SPDX-License-Identifier: EPL-2.0
 ********************************************************************************/

#pragma once

#include <ApiClassInterfacesV1_2.h>
#include <MantleAPI/Common/i_logger.h>
#include <MantleAPI/Execution/i_environment.h>
#include <MantleAPI/Execution/i_scenario_engine.h>
#include <openScenarioLib/src/common/SimpleMessageLogger.h>

#include <memory>
#include <string>

namespace OpenScenarioEngine::v1_2
{
class ControllerCreator;
class EntityCreator;
class IProbabilityService;

namespace Node
{
class RootNode;
}  // namespace Node

class OpenScenarioEngine : public mantle_api::IScenarioEngine
{
public:
  using OpenScenarioPtr = std::shared_ptr<NET_ASAM_OPENSCENARIO::v1_2::IOpenScenario>;
  using ScenarioDefinitionPtr = std::shared_ptr<NET_ASAM_OPENSCENARIO::v1_2::IScenarioDefinition>;
  using SimpleMessageLoggerPtr = std::shared_ptr<NET_ASAM_OPENSCENARIO::SimpleMessageLogger>;

  /// Creates an instance of the openSCENARIO Engine (OSE)
  ///
  /// @param scenario_file_path  path to the scenario description (*.xosc)
  /// @param environment         interface to the environment, which is controlled by the OSE
  /// @param logger              instance of a potential logger
  /// @param initial_seed        initial seed for internally used random generators (defaults to 0)
  OpenScenarioEngine(
      const std::string& scenario_file_path,
      std::shared_ptr<mantle_api::IEnvironment> environment,
      std::shared_ptr<mantle_api::ILogger> logger = nullptr,
      unsigned int initial_seed = 0);

  void Init() override;

  [[nodiscard]] mantle_api::ScenarioInfo GetScenarioInfo() const override;
  void Step() override;
  [[nodiscard]] bool IsFinished() const override;
  void ActivateExternalHostControl() override;
  int ValidateScenario() override;
  OpenScenarioPtr GetScenario() { return scenario_data_ptr_; }

protected:
  std::string scenario_file_path_;
  OpenScenarioPtr scenario_data_ptr_{nullptr};
  std::shared_ptr<NET_ASAM_OPENSCENARIO::SimpleMessageLogger> catalog_message_logger_;
  std::shared_ptr<NET_ASAM_OPENSCENARIO::SimpleMessageLogger> message_logger_;
  void LoadScenarioData();

private:
  /// If the path cannot be resolved by the scenario loader, override this function to resolve the path before loading
  /// in ParseScenarioFile().
  ///
  /// @param Scenario file in any given form
  /// @return Resolved path to the scenario file
  [[nodiscard]] virtual std::string ResolveScenarioPath(const std::string& scenario_path) const
  {
    return scenario_path;
  }

  /// If the map path cannot be resolved by the environment, override this function to resolve the path before calling
  /// the environment's CreateMap().
  ///
  /// @param Scenario file in any given form
  /// @return Resolved path to the scenario file
  [[nodiscard]] virtual std::string ResolveMapPath(const std::string& map_path) const
  {
    return map_path;
  }

  /// If a different standard output system is used, override this function with the proper output
  ///
  /// @param message_logger Where the messages are extracted from
  virtual void LogParsingMessages(SimpleMessageLoggerPtr message_logger);

  void ParseScenarioFile();
  void LoadRoadNetwork();
  void LoadUsedArea(std::shared_ptr<NET_ASAM_OPENSCENARIO::v1_2::IUsedArea> used_area,
                    mantle_api::MapDetails& map_details);
  void CreateEntities();
  void CreateControllers();
  void CreateAndInitStoryboard();

  /// The default environmental conditions are excellent, no precipitation, illumination as other,
  /// humidity at 60 percent, temperature at 15 degrees celsius and atmospheric pressure at sea level.
  /// The time of the day is considered as midday.
  /// The default routing behavior according to the standard is random.
  ///
  /// @return void
  void SetEnvironmentDefaults();

  void SetScenarioDefinitionPtr(OpenScenarioPtr open_scenario_ptr);
  [[nodiscard]] mantle_api::Time GetDuration() const;

  std::shared_ptr<mantle_api::IEnvironment> environment_;
  std::shared_ptr<EntityCreator> entity_creator_;
  std::shared_ptr<ControllerCreator> controller_creator_;
  std::shared_ptr<IProbabilityService> probability_service_;
  std::shared_ptr<mantle_api::ILogger> logger_{nullptr};
  std::shared_ptr<Node::RootNode> root_node_{nullptr};
  const mantle_api::ICoordConverter* coordinate_converter_{nullptr};
  ScenarioDefinitionPtr scenario_definition_ptr_{nullptr};

  bool finished_{false};
};
}  // namespace OpenScenarioEngine::v1_2
